import torch
from torch import nn


class VitClassTokens(nn.Module):
    def __init__(self, dim: int, num_tokens: int = 1):
        super().__init__()
        self.num_tokens = num_tokens
        if num_tokens > 0:
            self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
        else:
            self.tokens = None
        self.reset_parameters()

    def reset_parameters(self):
        if self.num_tokens > 0:
            nn.init.normal_(self.tokens, std=.02)

    def forward(self, x):
        if self.num_tokens == 0:
            return x
        assert x.ndim == 3
        tokens = self.tokens.expand(len(x), -1, -1)
        x = torch.concat([tokens, x], dim=1)
        return x

    def split(self, x):
        if self.num_tokens == 0:
            return None, x
        assert x.ndim == 3
        cls_tokens = x[:, :self.num_tokens]
        patch_tokens = x[:, self.num_tokens:]
        return cls_tokens, patch_tokens



import einops
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)


class PerceiverAttention(nn.Module):
    def __init__(self, dim, num_heads=8, bias=True, concat_query_to_kv=False, init_weights="truncnormal"):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.concat_query_to_kv = concat_query_to_kv
        self.init_weights = init_weights

        self.kv = nn.Linear(dim, dim * 2, bias=bias)
        self.q = nn.Linear(dim, dim, bias=bias)
        self.proj = nn.Linear(dim, dim, bias=bias)

        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            init_xavier_uniform_merged_linear(self.kv, num_layers=2)
        elif self.init_weights == "truncnormal":
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError

    def forward(self, q, kv, attn_mask=None):
        # project to attention space
        if self.concat_query_to_kv:
            kv = torch.concat([kv, q], dim=1)
        kv = self.kv(kv)
        q = self.q(q)

        # split per head
        q = einops.rearrange(
            q,
            "bs seqlen_q (num_heads head_dim) -> bs num_heads seqlen_q head_dim",
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        )
        k, v = einops.rearrange(
            kv,
            "bs seqlen_kv (two num_heads head_dim) -> two bs num_heads seqlen_kv head_dim",
            two=2,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)

        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x = self.proj(x)
        return x


class PerceiverAttention1d(PerceiverAttention):
    pass



from functools import partial

import torch

_cache = {}


def _wrapper(key, ctor=None):
    if key in _cache:
        tensor = _cache[key]
        # noinspection PyProtectedMember
        assert tensor._version == 0, f"cached tensor with key={key} was modified inplace"
        return tensor
    assert ctor is not None
    tensor = ctor()
    _cache[key] = tensor
    return tensor


def named(name, ctor=None):
    # force string to avoid any possibility of named overwriting other attributes in the cache
    # e.g. by passing ("zeros", (5,), "cpu", torch.bool) as name it would overwrite a cached zeros tensor
    assert isinstance(name, str)
    return _wrapper(key=name, ctor=ctor)


def zeros(size, device=None, dtype=None):
    return _wrapper(
        key=("zeros", size, str(device), dtype),
        ctor=partial(torch.zeros, size=size, device=device, dtype=dtype),
    )


def zeros_like(tensor):
    return zeros(size=tensor.shape, device=tensor.device, dtype=tensor.dtype)


def ones(size, device=None, dtype=None):
    return _wrapper(
        key=("ones", size, str(device), dtype),
        ctor=partial(torch.ones, size=size, device=device, dtype=dtype),
    )


def ones_like(tensor):
    return ones(size=tensor.shape, device=tensor.device, dtype=tensor.dtype)


def full(size, fill_value, device=None, dtype=None):
    return _wrapper(
        key=("full", size, fill_value, str(device), dtype),
        ctor=partial(torch.full, size=size, fill_value=fill_value, device=device, dtype=dtype),
    )


def full_like(tensor, fill_value):
    return full(size=tensor.shape, fill_value=fill_value, device=tensor.device, dtype=tensor.dtype)


def arange(start, end=None, step=1, device=None, dtype=None):
    key = ("arange", start, end, step, str(device), dtype)
    if end is None:
        ctor = partial(torch.arange, start=0, end=start, step=step, device=device, dtype=dtype)
    else:
        ctor = partial(torch.arange, start=start, end=end, step=step, device=device, dtype=dtype)
    return _wrapper(key=key, ctor=ctor)
